/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   Per-branch ACL implementation

*/

#define _BSD_SOURCE 1

#include "config.h"

#include "branch-acl.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


BRANCH_ACL *
find_branch_acl (DIR_ACL *dir_acl, char *branch)
{
    struct BRANCH_ACL *current = dir_acl->branch_acls;

    while (current != NULL) {
	if (strcmp(current->branch, branch) == 0)
	    return current;

	current = current->next;
    }

    return NULL;
}

BRANCH_ACL *
get_branch_acl (DIR_ACL *dir_acl, char *branch)
{
    struct BRANCH_ACL *result;

    load_branch_acls(dir_acl);

    result = find_branch_acl(dir_acl, branch);

    if (result == NULL)
	result = create_branch_acl(dir_acl, branch);

    return result;
}


BRANCH_ACL *
create_branch_acl (DIR_ACL *dir_acl, char *branch)
{
    struct BRANCH_ACL *branch_acl;

    branch_acl = calloc(1, sizeof(struct BRANCH_ACL));
    if (branch_acl == NULL)
	return NULL;

    branch_acl->branch = strdup(branch);
    if (branch_acl->branch == NULL) {
	free(branch_acl);
	return NULL;
    }

    branch_acl->dir_acl = dir_acl;

    branch_acl->next = dir_acl->branch_acls;
    dir_acl->branch_acls = branch_acl;

    return branch_acl;
}


BRANCH_ACE *
get_file_on_branch_ace (BRANCH_ACL *branch_acl, char *filename)
{
    BRANCH_ACE *current = branch_acl->first_ace;
    BRANCH_ACE *branch_ace = NULL;

    while (current != NULL) {
	if (strcmp(current->filename->buf, filename) == 0)
	    return current;

	current = current->next;
    }

    branch_ace = calloc(1, sizeof(struct BRANCH_ACE));
    if (branch_ace == NULL)
	return NULL;

    branch_ace->acl = branch_acl;

    branch_ace->filename = new_stringbuf(filename);
    if (branch_ace->filename == NULL) {
	free(branch_ace);
	return NULL;
    }

    branch_ace->fullpath = dup_stringbuf(branch_acl->dir_acl->dirname);
    if (branch_ace->fullpath == NULL) {
	free(branch_ace);
	return NULL;
    }
    catc_stringbuf(branch_ace->fullpath, '/');
    cat_stringbuf(branch_ace->fullpath, filename);

    branch_ace->next = branch_acl->first_ace;
    branch_acl->first_ace = branch_ace;

    return branch_ace;
}


static USER_ACE *
get_inherited_user_ace (struct DIR_ACL *dir_acl, char *branch, char *username)
{
    struct DIR_ACL **top_dirs = dir_acl->top_dirs;
    USER_ACE *last_user_ace = NULL;

    if (top_dirs != NULL) {
	for (; *top_dirs; top_dirs++) {
	    struct DIR_ACL *top_dir = *top_dirs;
	    struct USER_ACL *default_file_acl = get_default_file_user_acl(top_dir, branch);

	    USER_ACE *user_ace = get_user_ace(default_file_acl, username);
	    
	    if (user_ace != NULL)
		last_user_ace = user_ace;
	}
    }

    return last_user_ace;
}


static int
internal_branch_permission_granted(BRANCH_ACE *branch_ace, 
				   char *username,
				   branch_permission perm)
{
    USER_ACE *ace;

    /* no list of users, use permissions on trunk, and then default
       permissions for directory */
    if (branch_ace->user_acls == NULL) {
	struct BRANCH_ACL *branch_acl = branch_ace->acl;
	struct DIR_ACL *dir_acl = branch_acl->dir_acl;

	if (strcmp(branch_acl->branch, "trunk") != 0) {
	    struct BRANCH_ACL *trunk_acl = get_branch_acl(dir_acl, "trunk");
	    struct BRANCH_ACE *trunk_ace = get_file_on_branch_ace(trunk_acl, branch_ace->filename->buf);

	    if ((trunk_ace != NULL) && (trunk_ace->user_acls != NULL))
		return internal_branch_permission_granted(trunk_ace, username, perm);
	}

	ace = get_inherited_user_ace(dir_acl, branch_acl->branch, username);
	if (ace == NULL) {
	    struct USER_ACL *default_file_acl = get_default_file_user_acl(dir_acl, branch_acl->branch);
	    ace = get_user_ace(default_file_acl, username);
	}
	if (ace == NULL)
	    return default_branch_perm & perm;
	return ace->perm & perm;
    }

    ace = get_user_ace(branch_ace->user_acls, username);

    if (ace == NULL)
	return branch_perm_none;

    return ace->perm & perm;

}

int
branch_permission_granted (BRANCH_ACE *branch_ace, 
			   char *username,
			   branch_permission perm)
{
    DIR_ACL *dir_acl = branch_ace->acl->dir_acl;
    int mask_perm = branch_perm_all;
    int result_perm = branch_perm_none;

    /* check if the parent directory allows access */
    if (!dir_permission_granted(dir_acl, username, dir_perm_access))
	return branch_perm_none;

    if (!dir_permission_granted(dir_acl, username, dir_perm_modify))
	mask_perm = branch_perm_checkout;

    result_perm = internal_branch_permission_granted (branch_ace, username, perm);

    return result_perm & mask_perm;
}


int
set_branch_ace (BRANCH_ACE *branch_ace,
		char *username,
		branch_permission perm)
{
    branch_ace->user_acls = set_user_ace(branch_ace->user_acls, username, perm);

    return perm;
}

int
parse_branch_permissions (char *perm_str)
{

    if (strcmp(perm_str, "checkin") == 0)
	return branch_perm_checkin;

    if (strcmp(perm_str, "checkout") == 0)
	return branch_perm_checkout;

    if (strcmp(perm_str, "checkin,checkout") == 0)
	return branch_perm_checkin | branch_perm_checkout;

    if (strcmp(perm_str, "checkout,checkin") == 0)
	return branch_perm_checkin | branch_perm_checkout;

    if (strcmp(perm_str, "none") == 0)
	return branch_perm_none;

    return branch_perm_invalid;
}

char *
branch_permissions_string (int perm)
{
    if (perm == branch_perm_none)
	return "none";

    if (perm == branch_perm_checkin)
	return "checkin";

    if (perm == branch_perm_checkout)
	return "checkout";

    if (perm == (branch_perm_checkin | branch_perm_checkout))
	return "checkin,checkout";

    return "invalid";
}

