#define _BSD_SOURCE 1

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cvsgroup-internal.h"
#include "cvsgroup.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif

struct group_elem {
    struct cvsgroup *group;
    struct group_elem *next;
};

static struct group_elem *first_group_elem = NULL;

struct cvsgroup *
get_cvsgroup (char *group)
{
    struct group_elem *current = first_group_elem;

    while (current != NULL) {
	if (strcmp(current->group->groupname, group) == 0)
	    return current->group;

	current = current->next;
    }
    return NULL;
}

struct cvsgroup *
create_cvsgroup (char *group)
{
    struct cvsgroup *result;
    struct group_elem *group_elem;

    if (group == NULL)
	return NULL;

    if (group[0] != '%')
	return NULL;

    if (get_cvsgroup(group) != NULL)
	return NULL;

    result = calloc(1, sizeof(struct cvsgroup));
    if (result == NULL)
	return NULL;

    result->groupname = strdup(group);
    if (result->groupname == NULL) {
	free_cvsgroup(result);
	return NULL;
    }

    result->users = NULL;
    result->users_allocated = 0;
    result->users_count = 0;

    group_elem = calloc(1, sizeof(struct group_elem));
    if (group_elem == NULL) {
	free_cvsgroup(result);
	return NULL;
    }

    group_elem->group = result;
    group_elem->next = first_group_elem;
    first_group_elem = group_elem;

    return result;
}

void
free_cvsgroup (struct cvsgroup *cvsgroup)
{
    char **users;
    if (cvsgroup == NULL)
	return;

    free(cvsgroup->groupname);
    if (cvsgroup->users != NULL) {
	for (users = cvsgroup->users; *users; users++)
	    free(*users);
    }
    free(cvsgroup->users);
    cvsgroup->users = NULL;
    cvsgroup->users_allocated = 0;
    cvsgroup->users_count = 0;
    free(cvsgroup);
}


int
add_user_to_group (char *username, char *group)
{
    struct cvsgroup *cvsgroup;

    if (group[0] != '%')
	return 0;

    cvsgroup = get_cvsgroup(group);
    if (cvsgroup == NULL)
	cvsgroup = create_cvsgroup(group);

    if (cvsgroup == NULL)
	return 0;

    return internal_add_user_to_group(username, cvsgroup);
}

#define ALLOC_DELTA 8

int
internal_add_user_to_group (char *username, struct cvsgroup *cvsgroup)
{
    char **new_users;
    char *username_copy;

    if (internal_user_in_group(username, cvsgroup))
	return 0;

    if (cvsgroup->users_count + 1 >= cvsgroup->users_allocated) {
	new_users = realloc(cvsgroup->users,
			    (cvsgroup->users_allocated + ALLOC_DELTA) * sizeof(char *));
	if (new_users == NULL)
	    return 0;

	cvsgroup->users = new_users;
	cvsgroup->users_allocated += ALLOC_DELTA;
    }

    username_copy = strdup(username);
    if (username_copy == NULL)
	return 0;

    cvsgroup->users[cvsgroup->users_count++] = username_copy;
    cvsgroup->users[cvsgroup->users_count] = NULL;

    return 1;
}

void
clear_groups (void)
{
    struct group_elem *current = first_group_elem;

    while (current != NULL) {
	struct group_elem *next = current->next;
	free_cvsgroup(current->group);
	free(current);

	current = next;
    }

    first_group_elem = NULL;
}

#define DELIM " \t,"

static int
process_cvsgroup_line (char *line)
{
    char *p = line;
    char *groupname = p;
    char *username;
    struct cvsgroup *cvsgroup;
    char *ptrptr;

    while (*p && (*p != ':'))
	p++;

    if (!*p)
	return 0;

    *p++ = '\0';

    cvsgroup = get_cvsgroup(groupname);
    if (cvsgroup == NULL)
	cvsgroup = create_cvsgroup(groupname);
    if (cvsgroup == NULL)
	return 0;

    username = strtok_r(p, DELIM, &ptrptr);
    if (username == NULL)
	return 0;
    do {
	if (username[0] == '\0')
	    continue;

	if (!internal_add_user_to_group(username, cvsgroup))
	    return 0;

	username = strtok_r(NULL, DELIM, &ptrptr);
    } while (username != NULL);

    return 1;
}


#define MAX_LINE_SIZE 128

int
read_cvsgroup_file (char *filename)
{
    FILE *cvsgroup;
    char buf[MAX_LINE_SIZE];
    char *p;
    int lineno = 1;

    cvsgroup = fopen(filename, "r");
    if (cvsgroup == NULL)
	return -1;

    while ((p = fgets(buf, sizeof buf, cvsgroup)) != NULL) {
	int len = strlen(buf);

	if (buf[len - 1] != '\n') 
	    goto err;

	buf[len - 1] = '\0';

	if (buf[0] != '%')
	    continue;

	if (!process_cvsgroup_line(buf))
	    goto err;

	lineno++;
    }

    lineno = 0;

 err:
    fclose(cvsgroup);

    return lineno;
}


int
internal_user_in_group (char *username, struct cvsgroup *cvsgroup)
{
    char **users;
    if (cvsgroup->users == NULL)
	return 0;

    for (users = cvsgroup->users; *users; users++) {
	if (strcmp(*users, username) == 0)
	    return 1;
    }

    return 0;
}


int
user_in_group (char *username, char *group)
{
    struct cvsgroup *cvsgroup;

    cvsgroup = get_cvsgroup(group);
    if (cvsgroup == NULL)
	return 0;

    return internal_user_in_group(username, cvsgroup);
}
