/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   Per-directory ACL implementation

*/

#define _BSD_SOURCE 1

#include "config.h"

#include "dir-acl.h"
#include "branch-acl.h"
#include "user-acl.h"

/* Is there a better solution ?? */
#include "system.h"

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <sys/stat.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_IO_H
#include <io.h>
#endif

static char *toplevel_directory = "./";

void
set_toplevel_directory (char *new_tl_dir)
{
    if (new_tl_dir == NULL)
	return;

    if (new_tl_dir[strlen(new_tl_dir) - 1] != '/')
	return;

    toplevel_directory = new_tl_dir;
}


static enum DIR_PERM current_default_dir_perm = dir_perm_all;

void
set_default_dir_perm (enum DIR_PERM new_default_dir_perm)
{
    current_default_dir_perm = new_default_dir_perm;
}

enum DIR_PERM
get_default_dir_perm (void)
{
    return current_default_dir_perm;
}

static DIR_ACL *create_and_load_dir_acl (char *dir, char *branch);
static DIR_ACL *internal_get_dir_acl (char *dir, char *branch, 
				      struct DIR_ACL *the_dir_acl);

DIR_ACL *
get_dir_acl (char *dir, char *branch)
{
    struct DIR_ACL *dir_acl = create_and_load_dir_acl(dir, branch);
    char *dir_copy = strdup(dir);
    if (dir_copy == NULL)
	return NULL;

    internal_get_dir_acl(dir_copy, branch, dir_acl);

    free(dir_copy);

    return dir_acl;
}


#define ALLOC_DELTA 8

static int
add_dir_to_path(struct DIR_ACL *dir_acl, struct DIR_ACL *top_dir) {
    struct DIR_ACL **top_dirs = dir_acl->top_dirs;
    int dirs_count = dir_acl->top_dirs_count;
    int dirs_allocated = dir_acl->top_dirs_allocated;

    /* reallocate if needed */
    if (dirs_count + 1 >= dirs_allocated) {
	struct DIR_ACL **new_top_dirs = 
	    realloc(top_dirs, (dirs_allocated + ALLOC_DELTA) * sizeof(DIR_ACL*));

	if (new_top_dirs == NULL)
	    return 0;

	top_dirs = dir_acl->top_dirs = new_top_dirs;
	dir_acl->top_dirs_allocated += ALLOC_DELTA;
    }

    top_dirs[dirs_count++] = top_dir;
    top_dirs[dirs_count] = NULL;

    dir_acl->top_dirs_count = dirs_count;

    return 1;
}


static DIR_ACL *
internal_get_dir_acl (char *dir, char *branch, struct DIR_ACL *the_dir_acl)
{
    int toplevel_len = strlen(toplevel_directory);
    char *last_component;

    /* check that requested directory derives from top-level directory */
    if (strncmp(dir, toplevel_directory, toplevel_len) != 0)
	return NULL;

    if (the_dir_acl == NULL)
	return NULL;

    /* FIXME: somehow the parent directory gets added twice.  It is
       harmless, but needs investigation */

    last_component = strrchr(dir + toplevel_len, '/');
    if (last_component == NULL) {
	/* we are processing the top dir */
	struct DIR_ACL *topdir_acl = create_and_load_dir_acl(dir, branch);

	add_dir_to_path(the_dir_acl, topdir_acl);

    } else {
	/* we need to strip the last component and try to get its top dir */
	struct DIR_ACL *this_dir_acl;
	*last_component = '\0';
	this_dir_acl = create_and_load_dir_acl(dir, branch);
	add_dir_to_path(the_dir_acl, this_dir_acl);

	internal_get_dir_acl(dir, branch, the_dir_acl);

	*last_component = '/';
    }
    return the_dir_acl;
}



#define MAX_LINE_SIZE 128

static DIR_ACL *
create_and_load_dir_acl (char *dir, char *branch)
{
    FILE *acl;
    struct DIR_ACL *dir_acl;
    struct stat stat_buf;
    struct stringbuf *filename;
    char buf[MAX_LINE_SIZE];

    errno = 0;

    if (stat(dir, &stat_buf) == -1)
	return NULL;

    dir_acl = (DIR_ACL *) calloc(1, sizeof(struct DIR_ACL));
    if (dir_acl == NULL)
	return NULL;

    dir_acl->dirname = new_stringbuf(dir);
    if (dir_acl->dirname == NULL) {
	free(dir_acl);
	return NULL;
    }

    filename = dup_stringbuf(dir_acl->dirname);
    if (filename == NULL) {
	free(dir_acl);
	return NULL;
    }
    cat_stringbuf(filename, "/DIRACL");

    acl = fopen(filename->buf, "r");
    if (acl == NULL)
	goto out;

    while (fgets(buf, MAX_LINE_SIZE, acl) != NULL) {
	int len = strlen(buf);
	char *p = buf;
	char initial_char;
	char *branch;
	char *username;
	char *perm_string;

	if (buf[len - 1] != '\n') 
	    goto err;

	buf[len - 1] = '\0';
	initial_char = *p++;
	if ((initial_char != '/') && (initial_char != 'D'))
	    continue;

	branch = p;
	while (*p && (*p != '\t'))
	    p++;
	if (!*p)
	    goto err;

	*p++ = '\0';
	username = p;

	while (*p && (*p != '\t'))
	    p++;
	if (!*p)
	    goto err;

	*p++ = '\0';
	perm_string = p;

	while (*p && (*p != '\t'))
	    p++;

	if (*p)
	    goto err;

	if (initial_char == '/')
	    set_dir_ace(dir_acl, username, parse_dir_permissions(perm_string));
	if (initial_char == 'D')
	    set_default_file_acl(dir_acl, branch, username, parse_branch_permissions(perm_string));
    }

    fclose(acl);

 out:
    return dir_acl;

 err:
    free_stringbuf(filename);
    if (acl != NULL)
	fclose(acl);
    return NULL;
}


#define EXTRACT_NEXT_LINE(c)       \
	while (*p && (*p != '\t')) \
	    p++;                   \
                                   \
	if (*p == c)               \
	    goto out;              \
        *p = '\0';                 \
        p++;

int
load_branch_acls (struct DIR_ACL *dir_acl)
{
    FILE *acl = NULL;
    char buf[MAX_LINE_SIZE];
    char *p;
    struct stringbuf *acl_filename;
    int retval = -1;
    struct BRANCH_ACL *current_bacl = NULL;
    struct BRANCH_ACE *current_bace = NULL;

    if (dir_acl->branch_acls_loaded)
	return 0;

    /* try to open ACL file, if it exists */
    acl_filename = dup_stringbuf(dir_acl->dirname);
    if (acl_filename == NULL)
	goto out;
    catc_stringbuf(acl_filename, '/');
    cat_stringbuf(acl_filename, "ACL");

    acl = fopen(acl_filename->buf, "r");
    if (acl == NULL) {
	if (errno == ENOENT)
	    goto out;
	else {
	    retval = -1;
	    goto out;
	}
    }

    retval = 1;
    while ((p = fgets(buf, MAX_LINE_SIZE, acl)) != NULL) {
	int len = strlen(buf);
	char initial_char;
	char *filename;
	char *branch;
	char *username;
	char *rights;
	int perm;
	struct USER_ACL *new_acl;

	if (buf[len - 1] != '\n') 
	    goto out;

	buf[len - 1] = '\0';

	initial_char = *p++;
	if (initial_char != ' ')
	    continue;

	branch = p;
	EXTRACT_NEXT_LINE('\0');
	if (!current_bacl || (strcmp(current_bacl->branch, branch) != 0)) {
	    current_bacl = find_branch_acl(dir_acl, branch);
	    if (current_bacl == NULL)
		current_bacl = create_branch_acl(dir_acl, branch);

	    if (current_bacl == NULL)
		goto out;
	    current_bace = NULL;
	}

	filename = p;
	EXTRACT_NEXT_LINE('\0');
	if (!current_bace || (strcmp(current_bace->filename->buf, filename) != 0)) {
	    current_bace = get_file_on_branch_ace(current_bacl, filename);
	    if (current_bace == NULL)
		goto out;
	}

	username = p;
	EXTRACT_NEXT_LINE('\0');
	rights = p;
	EXTRACT_NEXT_LINE('\t');
	perm = parse_branch_permissions(rights);

	new_acl = set_user_ace(current_bace->user_acls, username, perm);
	if (new_acl == NULL) {
	    retval = -1;
	    goto out;
	}
	current_bace->user_acls = new_acl;

	retval++;
    }

    if (!feof(acl)) {
	retval = -1;
	goto out;
    }

    dir_acl->branch_acls_loaded = 1;

    retval = 0;

 out:
    if (acl != NULL)
	if (fclose(acl) == EOF)
	    retval = -1;

    free_stringbuf(acl_filename);
    return retval;

}


int
dir_permission_granted (struct DIR_ACL *dir_acl, char *username, int perm)
{
    struct DIR_ACL **top_dirs;
    USER_ACE* ace;
    enum DIR_PERM eff_perm = dir_perm_all;
    int user_was_mentioned_in_acl = 0;

    if (dir_acl->top_dirs) {
	for (top_dirs = dir_acl->top_dirs; *top_dirs; top_dirs++) {
	    struct DIR_ACL* parent_acl = *top_dirs;
	    USER_ACE* ace;

	    /* restrict current parent permissions */
	    ace = get_user_ace(parent_acl->user_acls, username);
	    if (ace != NULL) {
		user_was_mentioned_in_acl = 1;
		eff_perm &= ace->perm;
	    }
	}
    }

    ace = get_user_ace(dir_acl->user_acls, username);
    /* ...we have four cases... */
    if (!user_was_mentioned_in_acl) {
	if (ace == NULL)
	    return perm & get_default_dir_perm();
	else
	    return perm & ace->perm;
    } else {
	/* we have seen parent with ACL */
	if (ace == NULL)
	    return perm & eff_perm;
	else
	    return perm & eff_perm & ace->perm;
    }
}


int
set_dir_ace (struct DIR_ACL *dir_acl, char *username, int perm)
{
    dir_acl->user_acls = set_user_ace(dir_acl->user_acls, username, perm);

    return perm;
}


int parse_dir_permissions (char *perm_string)
{
    if (strcmp(perm_string, "access") == 0)
	return dir_perm_access;

    if (strcmp(perm_string, "modify") == 0)
	return dir_perm_modify;

    if (strcmp(perm_string, "access,modify") == 0)
	return dir_perm_all;

    if (strcmp(perm_string, "modify,access") == 0)
	return dir_perm_all;

    if (strcmp(perm_string, "none") == 0)
	return dir_perm_none;

    return dir_perm_invalid;
}

char *
dir_permissions_string (int perm)
{
    if (perm == dir_perm_none)
	return "none";

    if (perm == dir_perm_access)
	return "access";

    if (perm == dir_perm_modify)
	return "modify";

    if (perm == dir_perm_all)
	return "access,modify";

    return "invalid";
}


static int
internal_store_dir_acl (int fd, DIR_ACL *dir_acl)
{
    int retval = -1;
    struct stringbuf *line = NULL;
    int saved_len;
    struct USER_ACL *user_acl = dir_acl->user_acls;

    line = new_stringbuf("/");
    if (line == NULL)
	goto out;

    cat_stringbuf(line, "trunk");
    catc_stringbuf(line, '\t');
    saved_len = line->len;
    while (user_acl != NULL) {
	cat_stringbuf(line, user_acl->username);
	catc_stringbuf(line, '\t');
	cat_stringbuf(line, dir_permissions_string(user_acl->perm));
	catc_stringbuf(line, '\n');

	if (write(fd, line->buf, line->len) != line->len)
	    goto out;

	shrink_stringbuf(line, saved_len);

	user_acl = user_acl->next;
    }

    retval = 0;

 out:
    free_stringbuf(line);

    return retval;
}

static int
internal_store_default_file_acl (FILE *acl_file, DIR_ACL *dir_acl)
{
    struct DEFAULT_FILE_ACL *dfa = dir_acl->default_file_acl;

    while (dfa != NULL) {
	struct USER_ACL *user_acl = dfa->user_acl;
	while (user_acl != NULL) {
	    fprintf(acl_file, "D%s\t%s\t%s\n", dfa->branch, 
		    user_acl->username, branch_permissions_string(user_acl->perm));

	    user_acl = user_acl->next;
	}
	dfa = dfa->next;
    }

    return 0;
}

int
store_dir_acl (struct DIR_ACL *dir_acl)
{
    int fd;
    FILE *acl_file;
    int retval = -1;
    struct stringbuf *filename = NULL;
    struct stringbuf *tmpfilename = NULL;

    filename = dup_stringbuf(dir_acl->dirname);
    if (filename == NULL)
	goto out;

    cat_stringbuf(filename, "/DIRACL");

    tmpfilename = dup_stringbuf(dir_acl->dirname);
    if (tmpfilename == NULL)
	goto out;

    cat_stringbuf(tmpfilename, "/DIRACL.tmp");

    fd = open(tmpfilename->buf, O_CREAT | O_EXCL | O_WRONLY, 0660);
    if (fd == -1)
	goto out;

    acl_file = fdopen(fd, "w");
    if (acl_file == NULL)
	goto out;

    if (internal_store_dir_acl(fd, dir_acl) != 0)
	goto out;

    if (internal_store_default_file_acl(acl_file, dir_acl) != 0)
	goto out;

    if (fclose(acl_file) == EOF)
	goto out;

    if (CVS_RENAME(tmpfilename->buf, filename->buf) == -1)
	goto out;

    retval = 0;
 out:
    free_stringbuf(tmpfilename);

    return retval;
}

int
store_branch_acls (struct DIR_ACL *dir_acl)
{
    struct stringbuf *tmpfilename = NULL;
    struct stringbuf *filename = NULL;
    FILE *acl_file = NULL;
    int tmpfd = -1;
    struct BRANCH_ACL *branch_acl = dir_acl->branch_acls;
    int retval = 0;

    tmpfilename = dup_stringbuf(branch_acl->dir_acl->dirname);
    if (tmpfilename == NULL)
	goto out;

    cat_stringbuf(tmpfilename, "/ACL.tmp");

    filename = dup_stringbuf(tmpfilename);
    if (filename == NULL)
	goto out;
    shrink_stringbuf(filename, filename->len - strlen(".tmp"));

    tmpfd = open(tmpfilename->buf, O_CREAT | O_EXCL | O_WRONLY, 0660);
    if (tmpfd == -1)
	goto out;

    acl_file = fdopen(tmpfd, "w");
    if (acl_file == NULL)
	goto out;

    while (branch_acl != NULL) {
	struct BRANCH_ACE *current_ace = branch_acl->first_ace;

	while (current_ace != NULL) {
	    USER_ACL *user_acl = current_ace->user_acls;

	    while (user_acl != NULL) {
		fprintf(acl_file, " %s\t%s\t%s\t%s\n", 
			branch_acl->branch, current_ace->filename->buf, 
			user_acl->username, branch_permissions_string(user_acl->perm));

		user_acl = user_acl->next;
	    }

	    current_ace = current_ace->next;
	}

	branch_acl = branch_acl->next;
    }
    if (fclose(acl_file) == EOF) {
	acl_file = NULL;
	goto out;
    }
    acl_file = NULL;

    if (CVS_RENAME(tmpfilename->buf, filename->buf) == -1)
	goto out;

    retval = 1;

 out:
    if (acl_file != NULL)
	fclose(acl_file);
    if (tmpfilename)
	unlink(tmpfilename->buf);
    free_stringbuf(tmpfilename);
    free_stringbuf(filename);
    return retval;
}


static struct DEFAULT_FILE_ACL *
get_default_file_acl (DIR_ACL *dir_acl, char *branch)
{
    struct DEFAULT_FILE_ACL *current = dir_acl->default_file_acl;

    while (current != NULL) {
	if (strcmp(current->branch, branch) == 0)
	    return current;

	current = current->next;
    }

    return NULL;
}

struct USER_ACL *
get_default_file_user_acl (DIR_ACL *dir_acl, char *branch)
{
    struct DEFAULT_FILE_ACL *dfa = get_default_file_acl(dir_acl, branch);
    if (dfa == NULL)
	return NULL;

    return dfa->user_acl;
}

int
set_default_file_acl (DIR_ACL *dir_acl, char *branch, char *username, int perm)
{
    struct DEFAULT_FILE_ACL *dfa;
    struct USER_ACL *new_acls;

    dfa = get_default_file_acl(dir_acl, branch);
    if (dfa == NULL) {
	dfa = calloc(1, sizeof(struct DEFAULT_FILE_ACL));
	if (dfa == NULL)
	    return 0;

	dfa->branch = strdup(branch);
	if (dfa->branch == NULL) {
	    free(dfa);
	    return 0;
	}

	dfa->next = dir_acl->default_file_acl;

	dir_acl->default_file_acl = dfa;
    }

    new_acls = set_user_ace(dfa->user_acl, username, perm);
    if (new_acls == NULL)
	return 0;

    dfa->user_acl = new_acls;

    return perm;
}
