/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   Per-module ACL implementation

*/

#include "module-acl.h"
#include "repository.h"

#include <stdlib.h>
#include <string.h>

struct MODULE_ACL *
get_module_acl (struct REPOSITORY *repos, char *module)
{
    struct MODULE_ACL *result;

    result = find_module_acl(repos, module);
    if (result != NULL)
	return result;

    result = calloc(1, sizeof(struct MODULE_ACL));
    if (result == NULL)
	return NULL;

    result->repository = repos;
    result->module = strdup(module);
    if (result->module == NULL) {
	free(result);
	return NULL;
    }

    result->next = repos->first_module;
    repos->first_module = result;

    return result;
}

static void free_module_tag_ace (struct MODULE_TAG_ACE *mta);

void
free_module_acl (struct MODULE_ACL *macl)
{
    struct MODULE_TAG_ACE *mta;
    if (macl == NULL)
	return;

    mta = macl->tag_acl;
    while (mta != NULL) {
	struct MODULE_TAG_ACE *next = mta->next;
	free_module_tag_ace(mta);
	mta = next;
    }

    free(macl->module);
    free(macl);
}

static void
free_module_tag_ace (struct MODULE_TAG_ACE *mta)
{
    if (mta == NULL)
	return;

    free(mta->username);
    free(mta->prefix);
    free(mta);
}

static struct MODULE_TAG_ACE *
find_module_tag_ace (struct MODULE_ACL *mod_acl,
		     char *username, char *prefix, enum TAGGING_MODE mode)
{
    struct MODULE_TAG_ACE *result = mod_acl->tag_acl;

    while (result != NULL) {
	if ((strcmp(result->username, username) == 0) &&
	    (strcmp(result->prefix, prefix) == 0) &&
	    (result->mode == mode))
	{
	    return result;
	}

	result = result->next;
    }

    return NULL;
}


static int
internal_module_tagging_granted (struct MODULE_ACL *mod_acl,
				 char *username, char *tag, enum TAGGING_MODE mode)
{
    struct MODULE_TAG_ACE *current = mod_acl->tag_acl;

    if (current == NULL)
	return 1;

    while (current != NULL) {
	if ((strcmp(current->username, username) == 0) &&
	    (strncmp(current->prefix, tag, strlen(current->prefix)) == 0) &&
	    (current->mode == mode))
	    return 1;
	current = current->next;
    }

    return 0;
}


int
module_tagging_granted (struct MODULE_ACL *mod_acl,
			char *username, char *tag)
{
    return internal_module_tagging_granted(mod_acl, username, tag, TAG_TAGGING);
}


int
module_branching_granted (struct MODULE_ACL *mod_acl,
			  char *username, char *tag)
{
    return internal_module_tagging_granted(mod_acl, username, tag, BRANCH_TAGGING);
}


static int
internal_grant_module_tagging (struct MODULE_ACL *mod_acl,
			       char *username, char *prefix, enum TAGGING_MODE mode)
{
    struct MODULE_TAG_ACE *mta;

    if (find_module_tag_ace(mod_acl, username, prefix, mode) != NULL)
	return 1;

    mta = calloc(1, sizeof(struct MODULE_TAG_ACE));
    if (mta == NULL)
	return 0;

    mta->username = strdup(username);
    if (mta->username == NULL)
	goto err;

    mta->prefix = strdup(prefix);
    if (mta->prefix == NULL)
	goto err;

    mta->mode = mode;

    mta->next = mod_acl->tag_acl;
    if (mta->next != NULL)
	mta->next->prev = mta;
    mod_acl->tag_acl = mta;
    mta->prev = NULL;

    return 1;

 err:
    free_module_tag_ace(mta);
    return 0;
}

int
grant_module_tagging (struct MODULE_ACL *mod_acl,
		      char *username, char *prefix)
{
    return internal_grant_module_tagging(mod_acl, username, prefix, TAG_TAGGING);
}

int
grant_module_branching (struct MODULE_ACL *mod_acl,
			char *username, char *prefix)
{
    return internal_grant_module_tagging(mod_acl, username, prefix, BRANCH_TAGGING);
}


static int
internal_revoke_module_tagging (struct MODULE_ACL *mod_acl,
				char *username, char *prefix, enum TAGGING_MODE mode)
{
    struct MODULE_TAG_ACE *mta;

    mta = find_module_tag_ace(mod_acl, username, prefix, mode);

    if (mta == NULL)
	return 1;

    if (mta->prev != NULL)
	mta->prev->next = mta->next;

    if (mta->next != NULL)
	mta->next->prev = mta->prev;

    if (mod_acl->tag_acl == mta)
	mod_acl->tag_acl = mta->next;

    free_module_tag_ace(mta);

    return 1;
}

int
revoke_module_tagging (struct MODULE_ACL *mod_acl,
			 char *username, char *prefix)
{
    return internal_revoke_module_tagging(mod_acl, username, prefix, TAG_TAGGING);
}


int
revoke_module_branching (struct MODULE_ACL *mod_acl,
			 char *username, char *prefix)
{
    return internal_revoke_module_tagging(mod_acl, username, prefix, BRANCH_TAGGING);
}

