/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   ACL-related repository handling

*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "system.h"

#include "stringbuf.h"
#include "repository.h"

#include <errno.h>
#include <stdio.h>
#ifdef HAVE_STDLIB_H 
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#define CVSROOT_MODULE_ACL "/CVSROOT/module-acl"

#define EXTRACT_NEXT_FIELD(c)      \
        while (*p && (*p != '\t')) \
            p++;                   \
                                   \
        if (*p == c)               \
            goto err;              \
        *p = '\0';                 \
        p++;


static int
internal_load_module_acl(struct REPOSITORY *repos)
{
    struct stringbuf *filename;
    FILE *acl_file = NULL;
    char line[256];
    char *p;
    int retval = -1;
    struct MODULE_ACL *macl = NULL;

    filename = new_stringbuf(repos->directory);
    if (filename == NULL)
	goto err;

    cat_stringbuf(filename, CVSROOT_MODULE_ACL);

    acl_file = fopen(filename->buf, "r");
    if (acl_file == NULL) {
	if (errno == ENOENT)
	    retval = 0;
	goto err;
    }

    retval = 1;
    while ((p = fgets(line, sizeof(line), acl_file)) != NULL) {
	int len = strlen(line);
	char *module;
	char *username;
	char *command;
	char *prefix;

	if (line[len - 1] != '\n')
	    goto err;

	line[len - 1] = '\0';

	module = p;
	EXTRACT_NEXT_FIELD('\0');

	username = p;
	EXTRACT_NEXT_FIELD('\0');

	command = p;
	EXTRACT_NEXT_FIELD('\0');

	prefix = p;
	EXTRACT_NEXT_FIELD('\t');

	if ((macl == NULL) || (strcmp(macl->module, module) != 0))
	    macl = get_module_acl(repos, module);

	if (strcmp(command, "tag") == 0) {
	    if (!grant_module_tagging(macl, username, prefix))
		goto err;
	} else if (strcmp(command, "branch") == 0) {
	    if (!grant_module_branching(macl, username, prefix))
		goto err;
	}

    }

    retval = 0;

 err:
    if (acl_file != NULL)
	fclose(acl_file);
    free_stringbuf(filename);
    return retval;
}

struct REPOSITORY *
get_repository (char *directory)
{
    struct REPOSITORY *result;

    result = calloc(1, sizeof(struct REPOSITORY));
    if (result == NULL)
	return NULL;

    result->directory = strdup(directory);
    if (result->directory == NULL) {
	free(result);
	return NULL;
    }

    if (internal_load_module_acl(result) != 0) {
	free(result);
	return NULL;
    }

    return result;
}

void
free_repository (struct REPOSITORY *repos)
{
    struct MODULE_ACL *macl;

    if (repos == NULL)
	return;

    macl = repos->first_module;
    while (macl != NULL) {
	struct MODULE_ACL *next = macl->next;
	free_module_acl(macl);
	macl = next;
    }

    free(repos->directory);
    free(repos);
}


struct MODULE_ACL *
find_module_acl (struct REPOSITORY *repos, char *module)
{
    struct MODULE_ACL *current = repos->first_module;

    while (current != NULL) {
	if (strcmp(current->module, module) == 0)
	    return current;

	current = current->next;
    }

    return NULL;
}



int
save_module_acl (struct REPOSITORY *repos)
{
    struct MODULE_ACL *current_module;
    struct stringbuf *filename = NULL;
    struct stringbuf *tmp_filename = NULL;
    FILE *acl_file = NULL;
    int retval = 0;

    filename = new_stringbuf(repos->directory);
    if (filename == NULL)
	goto out;

    cat_stringbuf(filename, CVSROOT_MODULE_ACL);
    tmp_filename = dup_stringbuf(filename);
    if (tmp_filename == NULL)
	goto out;

    cat_stringbuf(tmp_filename, ".tmp");

    acl_file = fopen(tmp_filename->buf, "w");
    if (acl_file == NULL)
	goto out;

    current_module = repos->first_module;
    while (current_module != NULL) {
	struct MODULE_TAG_ACE *current = current_module->tag_acl;
	while (current != NULL) {
	    char *command;
	    if (current->mode == TAG_TAGGING)
		command = "tag";
	    else if (current->mode == BRANCH_TAGGING)
		command = "branch";
	    else
		command = "error";
	    fprintf(acl_file, "%s\t%s\t%s\t%s\n",
		    current_module->module, current->username, 
		    command, current->prefix);
		    
	    current = current->next;
	}
	current_module = current_module->next;
    }

    if (fclose(acl_file) == EOF) {
	acl_file = NULL;
	goto out;
    }
    acl_file = NULL;

    if (CVS_RENAME(tmp_filename->buf, filename->buf) == -1)
	goto out;

    retval = 1;

 out:
    if (acl_file != NULL)
	fclose(acl_file);
    free_stringbuf(filename);
    free_stringbuf(tmp_filename);
    return retval;
}
