/* 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.  
   
   Copyright (C) Alexey Mahotkin <alexm@hsys.msk.ru> 1999-2000

   Authenticates user against CVSROOT/cvspasswd file in repository.
   Complies to checkpassword protocol: http://cr.yp.to/checkpwd.html

*/

#define _BSD_SOURCE 1

#include "config.h"

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "env.h"

#define BUFFER_SIZE 256
#define CVSROOT_CVSPASSWD "/CVSROOT/cvspasswd"

char up[513];
int uplen;

char *err_prefix = NULL;

int main(int argc, char **argv)
{
    int anonymous;
    int auth_ok = 0;
    char *buffer;
    char *cvspasswd;
    FILE *CVSPASSWD;
    int i;
    char *field_user, *field_password, *field_system_user = NULL;
    char *field_email, *field_gecos, *field_status;
    int r;
    char *username;
    char *password;
    char *repository;
    char *system_user;
    
    if (!argv[1])
	exit (2);

    if (strcmp (getenv("CVSSERVER"), "nserver") == 0)
	err_prefix = "550 ";
    else
	if (strcmp (getenv("CVSSERVER"), "pserver") == 0)
	    err_prefix = "E ";

    uplen = 0;
    for (;;) {
	do
	    r = read(3, up + uplen, sizeof(up) - uplen);
	while ((r == -1) && (errno == EINTR));
	if (r == -1) 
	    _exit(111);
	if (r == 0) 
	    break;
	uplen += r;
	if (uplen >= sizeof(up)) 
	    _exit(1);
    }

    close(3);

    i = 0;
    username = up + i;
    while (up[i++]) if (i == uplen) _exit(2);
    password = up + i;
    if (i == uplen) _exit(2);
    while (up[i++]) if (i == uplen) _exit(2);
    repository = up + i;
    if (i == uplen) _exit(2);
    while(up[i++]) if (i == uplen) _exit(2);

#define OUT_OF_MEMORY					\
{							\
    fputs(err_prefix, stderr);				\
    fprintf(stderr, "cvschkpw: out of memory\n");	\
    _exit(111);						\
}							\

    /* authenticate against CVSROOT/cvspasswd */
    anonymous = (strcmp(username, "anonymous") == 0);
    buffer = malloc(BUFFER_SIZE);
    if (!buffer)
	OUT_OF_MEMORY;
    
    cvspasswd = malloc(strlen(repository) + strlen(CVSROOT_CVSPASSWD) + 1);
    if (!cvspasswd)
	OUT_OF_MEMORY;
    
    strcpy(cvspasswd, repository);
    strcat(cvspasswd, CVSROOT_CVSPASSWD);
    
    CVSPASSWD = fopen(cvspasswd, "r");
    if (!CVSPASSWD) {
	fputs(err_prefix, stderr);
	fprintf(stderr, "cvschkpw: could not open %s: %s\n", cvspasswd, strerror(errno));
	_exit(111);
    }
    
#define EXTRACT_NEXT_FIELD(field1, field2)	\
field2 = strchr(field1, ':');			\
if (!field2) {					\
    memset(buffer, '\0', BUFFER_SIZE);		\
    _exit(111);					\
}						\
						\
*field2++ = '\0';				\


    while (fgets(buffer, BUFFER_SIZE, CVSPASSWD))
    {
	field_user = buffer;
	EXTRACT_NEXT_FIELD(field_user, field_password);
	EXTRACT_NEXT_FIELD(field_password, field_system_user);

	if (strcmp(username, field_user) != 0)
	    continue;
	
	if (anonymous)
	    auth_ok = 1;
	else 
	    if (strcmp(crypt(password, field_password), field_password) == 0)
		auth_ok = 1;
	break;
    }

    memset(up, '\0', sizeof(up));
    memset(field_password, '\0', strlen(field_password));
    
    if (!auth_ok)
	_exit(1);
    
    EXTRACT_NEXT_FIELD(field_system_user, field_email);
    EXTRACT_NEXT_FIELD(field_email, field_gecos);
    EXTRACT_NEXT_FIELD(field_gecos, field_status);
    
    system_user = strdup(field_system_user);
    if (!system_user)
	OUT_OF_MEMORY;
    
    memset(buffer, '\0', BUFFER_SIZE);
    free(buffer);
    
    fclose(CVSPASSWD);
    free(cvspasswd);
    
    /* switch to specified user and group */
    {
	struct group *group = NULL; 
	struct passwd *passwd = NULL;
	
	/* switch to specified group */
	char *system_group = strchr(system_user, '.');
	if (system_group)
	{
	    *system_group++ = '\0';
	    
	    group = getgrnam(system_group);
	    if (!group) 
	    {
		fputs (err_prefix, stderr);
		printf("cvschkpw: could not find group %s: %s\n", 
		       system_group, strerror(errno));
		fflush(stdout);
		exit(EXIT_FAILURE);
	    }

	    if (group->gr_gid == 0)
	    {
		fputs (err_prefix, stderr);
		printf("cvschkpw: I will not setgid(0)\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	    }
	    if (setgid(group->gr_gid) != 0) 
	    {
		fputs (err_prefix, stderr);
		printf("cvschkpw: could not switch to group %s: %s\n", 
		       system_group, strerror(errno));
		fflush(stdout);
		exit(EXIT_FAILURE);
	    }
	}
	
	/* switch to specified user */
	passwd = getpwnam(system_user);
	if (!passwd) 
	{
	    fputs(err_prefix, stderr);
	    printf("cvschkpw: could not find user %s: %s\n",
		   system_user, strerror(errno));
	    fflush(stdout);
	    exit(EXIT_FAILURE);
	}
	if (passwd->pw_uid == 0) 
	{
	    fputs(err_prefix, stderr);
	    printf("cvschkpw: I will not setuid(0)\n");
	    fflush(stdout);
	    exit(EXIT_FAILURE);
	}
	
#ifdef HAVE_INITGROUPS
	if (getuid()==0 && initgroups(system_user, passwd->pw_gid) != 0) 
	{
	    fputs(err_prefix, stderr);
	    printf("cvschkpw: could not set groups for user %s: %s\n",
		   system_user, strerror(errno));
	    fflush(stdout);
	    exit(EXIT_FAILURE);
	}
#endif
	/* If group was not specified in cvspasswd, switch to 
	   user's primary group */
	if (!group) 
	{
	    if (setgid(passwd->pw_gid) != 0) 
	    {
		fputs(err_prefix, stderr);
		printf("cvschkpw: could not switch to group %d: %s\n",
		       passwd->pw_gid, strerror(errno));
		fflush(stdout);
		exit(EXIT_FAILURE);
	      }
	}
	if (setuid(passwd->pw_uid) != 0)
	{
	    fputs(err_prefix, stderr);
	    printf("cvschkpw: could not switch to user %s: %s\n",
		   system_user, strerror(errno));
	    fflush(stdout);
	    exit(EXIT_FAILURE);
	}
	if (chdir(passwd->pw_dir) == -1) {
	    fputs(err_prefix, stderr);
	    fprintf(stderr, "cvschkpw: could not chdir(%s): %s\n", 
		    passwd->pw_dir, strerror(errno));
	    _exit(111);
	}
	setenvvar("USER", passwd->pw_name);
	setenvvar("HOME", passwd->pw_dir);
	setenvvar("SHELL", passwd->pw_shell);
    }

    execvp(argv[1], argv + 1);
    fputs(err_prefix, stderr);
    fprintf(stderr, "cvschkpw: could not execvp(%s): %s\n", argv[1], strerror(errno));
    _exit(111);
}
