/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 1999-2000

   "cvspasswd" lets ordinary users change their passwords in
   CVSROOT/cvspasswd file.  This program is intended to be installed
   setuid "cvsadmin" user, who owns the CVSROOT/cvspasswd file.

   Changing passwords is allowed iff @file{CVSROOT/allow-cvspasswd} exists.

*/

#define _BSD_SOURCE 1

#include "config.h"

#include <assert.h>
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#include <stdio.h>
#include <stdlib.h>

#include "getline.h"

#include "cvspasswd-file.h"
#include "gensalt.h"
#include "lockpasswd.h"

static char *unlink_filename = NULL;

void atexit_unlink_filename PROTO((void));

void 
atexit_unlink_filename(void)
{
    if (unlink_filename)
	if (unlink(unlink_filename) == -1)
	    fprintf(stderr, "Cannot unlink() %s: %s\n", unlink_filename, strerror(errno));

}

#define OUT_OF_MEMORY				\
    {						\
	fprintf(stderr, "Out of memory\n");	\
	exit(EXIT_FAILURE);			\
    }

#define ALLOWED "/CVSROOT/allow-cvspasswd"

static int
allow_cvspasswd PROTO((char *repos));

static int
allow_cvspasswd (repos)
    char *repos;
{
    char *allowed = malloc(strlen(repos) + strlen(ALLOWED) + 1);
    struct stat buf;
    int result;

    strcpy(allowed, repos);
    strcat(allowed, ALLOWED);

    result = stat(allowed, &buf) == 0;
    free(allowed);

    return result;
}

#define BUF_SIZE 512
static char passwords[BUF_SIZE];

int
main(int argc, char *argv[])
{
    int auth_ok = 0;
    struct stat buf;
    char *newfilename, *oldfilename;
    FILE *CVSPASSWD, *NEWCVSPASSWD;
    char *CVSROOT;
    char *field_user, *field_password, *field_remainder;
    char *new_password;
    char *old_password = NULL;
    char *passwd_line = NULL;
    unsigned int passwd_line_allocated = 0;
    char *username;
    char *repository;
    char *salt;
    int working_as_admin = 0;

    repository = getenv("REPOSITORY");
    if (!repository) 
    {
	printf("450 %s: REPOSITORY environment variable not defined\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    username = getenv("CVSUSER");
    if (!username)
    {
 	printf("450 %s: CVSUSER environment variable not defined\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    srand (time (0));

    CVSROOT = malloc(strlen(repository) + strlen("/CVSROOT") + 1);
    if (!CVSROOT)
	OUT_OF_MEMORY;

    strcpy(CVSROOT, repository);
    strcat(CVSROOT, "/CVSROOT");

    if (stat(CVSROOT, &buf) != 0)
    {
	printf("450 %s: Cannot stat %s: %s\n", 
	       argv[0], CVSROOT, strerror(errno));
	exit(EXIT_FAILURE);
    }

    if (buf.st_uid != geteuid())
    {
	printf("450 %s: %s must belong to owner of %s\n", 
	       argv[0], CVSROOT, argv[0]);
	exit(EXIT_FAILURE);
    }

    if ((buf.st_mode & (S_IWGRP | S_IWOTH)) != 0)
    {
	printf("450 %s: %s must not have group-write or others-write bit set\n", 
	       argv[0], CVSROOT);
	exit(EXIT_FAILURE);
    }

    /* ordinary users must enter old password to change it.  
       "cvsadmin" user is allowed to set password directly */
    working_as_admin = (buf.st_uid == getuid());

    if (!working_as_admin && !allow_cvspasswd(repository)) {
 	printf("450 %s: you are not allowed to change passwords in repository\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    read(3, &passwords, BUF_SIZE - 1);
    passwords[BUF_SIZE - 1] = '\0';
    old_password = passwords;
    if (strlen(old_password) == (BUF_SIZE - 1)) {
	printf("450 %s: invalid user or password\n", argv[0]);
	exit(EXIT_FAILURE);
    }
    new_password = old_password + strlen(old_password) + 1;

    oldfilename = malloc(strlen(CVSROOT) + strlen("/cvspasswd") + 1);
    if (!oldfilename)
	OUT_OF_MEMORY;

    strcpy(oldfilename, CVSROOT);
    strcat(oldfilename, "/cvspasswd");

    CVSPASSWD = fopen(oldfilename, "r");
    if (!CVSPASSWD)
    {
	printf("450 Cannot open %s: %s\n", oldfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* first check user and/or password */

    while (cvs_getline (&passwd_line, &passwd_line_allocated, CVSPASSWD) != -1)
    {
	assert(strlen(passwd_line) > 0);
	passwd_line[strlen(passwd_line) - 1] = '\0';

	field_user = strdup(passwd_line);
	if (!field_user)
	    OUT_OF_MEMORY;

#define EXTRACT_NEXT_FIELD(field1, field2)		\
field2 = strchr(field1, ':');				\
if (!field2)						\
    continue;						\
							\
*field2++ = '\0';					\


	EXTRACT_NEXT_FIELD(field_user, field_password);
	if (strcmp(field_user, username) != 0)
	    continue;

	/* check old password only if running as ordinary user */
	if (!working_as_admin) {
	    EXTRACT_NEXT_FIELD(field_password, field_remainder);
	    auth_ok = (strcmp(crypt(old_password, field_password), field_password) == 0);
	} else
	    auth_ok = 1;

	break;
    }

    if (fclose(CVSPASSWD) == EOF)
    {
	printf("450 %s: cannot close %s: %s\n", 
	       argv[0], oldfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }
    
    if (!auth_ok)
    {
	if (working_as_admin)
	    printf("450 %s: there is no user %s in %s\n", argv[0], username, oldfilename);
	else
	    printf("450 %s: invalid username or password\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    /* and now we can lock cvspasswd */
    lockpasswd(repository);

    CVSPASSWD = fopen(oldfilename, "r");
    if (!CVSPASSWD)
    {
	printf("450 %s: cannot open %s: %s\n", argv[0], oldfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }

    newfilename = malloc(strlen(CVSROOT) + strlen("/cvspasswd.new") + 1);
    if (!newfilename)
	OUT_OF_MEMORY;

    strcpy(newfilename, CVSROOT);
    strcat(newfilename, "/cvspasswd.new");

    unlink_filename = newfilename;
    atexit(atexit_unlink_filename);

    NEWCVSPASSWD = create_cvspasswd(newfilename);
    if (!NEWCVSPASSWD)
    {
	printf("450 %s: cannot create %s: %s\n", 
	       argv[0], newfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }

    field_user = find_user (CVSPASSWD, NEWCVSPASSWD, username, find_action_return);

    if (!field_user)
    {
	printf("450 %s: Whoa!  User %s mysteriously disappeared from %s\n", 
		argv[0], username, oldfilename);
	exit(EXIT_FAILURE);
    }

#undef EXTRACT_NEXT_FIELD

#define EXTRACT_NEXT_FIELD(field1, field2)			\
field2 = strchr(field1, ':');					\
if (!field2)							\
{								\
    printf("450 %s: invalid line in %s for user %s\n",		\
           argv[0], oldfilename, username);			\
    exit(EXIT_FAILURE);						\
}								\
*field2++ = '\0';						\

    EXTRACT_NEXT_FIELD(field_user, field_password);
    EXTRACT_NEXT_FIELD(field_password, field_remainder);

    salt = gensalt ();
    if (!salt)
	OUT_OF_MEMORY;

    fprintf(NEWCVSPASSWD, "%s:%s:%s\n", 
	    field_user, crypt(new_password, salt), field_remainder);

    free(salt);

    /* copy the rest of cvspasswd to cvspasswd.new */
    find_user (CVSPASSWD, NEWCVSPASSWD, username, find_action_do_none);

    if (fclose(CVSPASSWD) == EOF)
    {
	printf("450 %s: cannot close %s: %s\n", argv[0], oldfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }
    
    if (fclose(NEWCVSPASSWD) == EOF)
    {
	printf("450 %s: cannot close %s: %s\n", argv[0], newfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }

    if (rename(newfilename, oldfilename) == -1)
    {
	printf("450 %s: cannot rename %s to %s: %s\n", 
	       argv[0], newfilename, oldfilename, strerror(errno));
	exit(EXIT_FAILURE);
    }

    unlink_filename = NULL;

    unlockpasswd(repository);

    exit(0);
}
