/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   Wrapper network client providing CVS protocol logging

*/

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "network-client.h"
#include "logging-client.h"

struct logging_client_data {
    char *logfile_name;
    FILE *log_in;
    FILE *log_out;
    FILE *log;
};

typedef struct logging_client_data *LCD;

static int logging_open(NCLI);
static int logging_close(NCLI);

static int logging_read(NCLI, char *buf, unsigned int len);
static int logging_write(NCLI, char *buf, unsigned int len);
static int logging_flush(NCLI);

struct network_client *
create_logging_client (cvsroot_t *CVSROOT, ncli_error_cb error_cb)
{
    LCD lcd;
    struct network_client *nc = malloc(sizeof(struct network_client));
    if (nc == NULL)
	return NULL;

    lcd = malloc(sizeof(struct logging_client_data));
    if (lcd == NULL)
	return NULL;

    nc->connect = logging_open;
    nc->disconnect = logging_close;

    nc->read = logging_read;
    nc->write = logging_write;
    nc->flush = logging_flush;

    nc->error = error_cb;

    nc->client_data = lcd;

    return nc;
}


/*
 *
 * Primary (new-style) log file name is getenv(CVS_CLIENT_LOG);
 * log-file of outgoing traffic is at CVS_CLIENT_LOG.out;
 * log-file of incoming traffic is at CVS_CLIENT_LOG.in
 *
 */ 
static int
logging_open (NCLI ncli)
{
    LCD lcd = (LCD) ncli->client_data;

    char *logfile_in_name;
    char *logfile_out_name;
    FILE *logfile;
    FILE *logfile_in;
    FILE *logfile_out;

    /* Get primary (new-style) log file name and open it */
    char *logfile_name = getenv("CVS_CLIENT_LOG");
    if (logfile_name == NULL)
	return ncli->error(ncli, "Internal error: environment variable CVS_CLIENT_LOG is not defined");

    lcd->logfile_name = logfile_name;

    logfile = fopen(logfile_name, "w");
    if (logfile == NULL)
	return ncli->error(ncli, "Cannot open %s for writing: %s", 
			   logfile_name, strerror(errno));

    setvbuf(logfile, NULL, _IONBF, 0);

    lcd->log = logfile;

    /* Get name of log file for incoming traffic and create this file */
    logfile_in_name = malloc(strlen(logfile_name) + strlen(".in") + 1);
    if (logfile_in_name == NULL)
	return ncli->error(ncli, "Out of memory");

    logfile_in_name = strcpy(logfile_in_name, logfile_name);
    strcat(logfile_in_name, ".in");

    logfile_in = fopen(logfile_in_name, "wb");
    if (logfile_in == NULL)
	return ncli->error(ncli, "Cannot open %s for writing: %s",
			   logfile_in_name, strerror(errno));

    free(logfile_in_name);
    setvbuf(logfile_in, NULL, _IONBF, 0);
    lcd->log_in = logfile_in;

    /* Get name of log file for outgoing traffic and create this file */
    logfile_out_name = malloc(strlen(logfile_name) + strlen(".out") + 1);
    if (logfile_out_name == NULL)
	return ncli->error(ncli, "Out of memory");

    strcpy(logfile_out_name, logfile_name);
    strcat(logfile_out_name, ".out");

    logfile_out = fopen(logfile_out_name, "wb");
    if (logfile_out == NULL)
	return ncli->error(ncli, "Cannot open %s for writing: %s",
			   logfile_out_name, strerror(errno));

    free(logfile_out_name);
    setvbuf(logfile_out, NULL, _IONBF, 0);
    lcd->log_out = logfile_out;

    return NCLI_SUCCESS;
}


/*
 *
 * Close all three log files, and call underlying ncli's disconnect
 *
 */

static int
logging_close (NCLI ncli)
{
    LCD lcd = (LCD) ncli->client_data;

    if (fclose(lcd->log) == EOF)
	return ncli->error(ncli, "Cannot close %s: %s",
			   lcd->logfile_name, strerror(errno));

    if (fclose(lcd->log_in) == EOF)
	return ncli->error(ncli, "Cannot close %s.in: %s",
			   lcd->logfile_name, strerror(errno));

    if (fclose(lcd->log_out) == EOF)
	return ncli->error(ncli, "Cannot close %s.out: %s",
			   lcd->logfile_name, strerror(errno));

    return ncli->next->disconnect(ncli->next);
}


static int
internal_log_line (char *arrow, FILE *stream, char *buf, unsigned int len)
{
    unsigned int i;

    if (len == 0)
	return 1;

    for (i = 0; i < len; i++) {
	if (fputs(arrow, stream) == EOF)
	    return 0;

	while (i < len) {
	    char ch = buf[i];
	    if (ch == '\n') {
		if (putc('\n', stream) == EOF)
		    return 0;
		break;
	    }
	    if (isprint(ch)) {
		if (putc(ch, stream) == EOF)
		    return 0;
	    } else {
		if (fprintf(stream, "\\%3o", ch) == EOF)
		    return 0;
	    }
	    i++;
	}
    }
    /* mark non-complete last line with plus sign */
    if (buf[len - 1] != '\n')
	fputs("+\n", stream);

    return 1;
}

/*
 *
 * Call underlying ncli for data, and write incoming traffic to log
 * files.
 *
 */

static int
logging_read (NCLI ncli, char *buf, unsigned int len)
{
    LCD lcd = ncli->client_data;
    int read = ncli->next->read(ncli->next, buf, len);

    if (read > 0) {
	if (fwrite(buf, 1, read, lcd->log_in) < (unsigned) read)
	    return ncli->error(ncli, "Cannot write to %s.in: (%d)%s", 
			       lcd->logfile_name, errno, strerror(errno));

	if (!internal_log_line("< ", lcd->log, buf, len))
	    return ncli->error(ncli, "Cannot write to %s: %s",
			       lcd->logfile_name, strerror(errno));
    }

    return read;
}

/*
 *
 * Write outgoing traffic to log files and pass control to underlying
 * ncli
 *
 */

static int
logging_write (NCLI ncli, char *buf, unsigned int len)
{
    LCD lcd = ncli->client_data;

    if (fwrite(buf, 1, len, lcd->log_out) < len)
	return ncli->error(ncli, "Cannot write to %s.out: %s", 
			   lcd->logfile_name, strerror(errno));

    if (!internal_log_line("> ", lcd->log, buf, len))
	return ncli->error(ncli, "Cannot write to %s: %s",
			   lcd->logfile_name, strerror(errno));

    return ncli->next->write(ncli->next, buf, len);
}


static int
logging_flush (NCLI ncli)
{
    return ncli->next->flush(ncli->next);
}
