/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   "lsacl" subcommand lists ACLs for specified directory in repository

*/

#include "config.h"

#include "cvs.h"
#include "acl.h"
#include "lsacl.h"
#include "recurse.h"

#include <stdlib.h>

static int local_only = 1;
static int deep = 0;
static size_t skip_prefix = 0;
static char *tagname = NULL;


static const char *const lsacl_usage[] =
{
    "Usage: %s %s subdirectories...\n",
    "\t-R\tRecursively list ACLs.\n",
    NULL
};


static struct stringbuf *
get_fulldir (char* relative_name)
{
    struct stringbuf *fulldir = new_stringbuf(current_parsed_root->directory);
    if (!fulldir)
	return NULL;

    catc_stringbuf(fulldir, '/');
    cat_stringbuf(fulldir, relative_name);

    return fulldir;
}

static void
forbidden_in (struct stringbuf* line, char* what, struct stringbuf* dirname)
{
    int saved_len = line->len;
    cat_stringbuf(line, what);
    cat_stringbuf(line, " forbidden in ");
    cat_stringbuf(line, dirname->buf);
    catc_stringbuf(line, '\n');
    cvs_output(line->buf, line->len);
    shrink_stringbuf(line, saved_len);
}

static void
forbidden (struct stringbuf* line, char* what)
{
    int saved_len = line->len;
    cat_stringbuf(line, what);
    cat_stringbuf(line, " forbidden\n");
    cvs_output(line->buf, line->len);
    shrink_stringbuf(line, saved_len);
}

enum topdirs_what { TOPDIRS_DIRECTORY, TOPDIRS_FILE };

static void
print_out_topdir_acls (DIR_ACL** topdir_acls, enum topdirs_what what, char* filename, struct stringbuf* line)
{
    if (topdir_acls) {
	for (; *topdir_acls; topdir_acls++) {
	    DIR_ACL* topdir_acl = *topdir_acls;
	    USER_ACL* user_acl = topdir_acl->user_acls;

	    while (user_acl) {
		int saved_len = line->len;
		if (filename) {
		    cat_stringbuf(line, filename);
		    catc_stringbuf(line, '/');
		}

		cat_stringbuf(line, user_acl->username);
		catc_stringbuf(line, '/');
		cat_stringbuf(line, "parent/");
		switch (what) {
		case TOPDIRS_DIRECTORY:
		    if ((user_acl->perm & dir_perm_access) != dir_perm_access)
			forbidden_in(line, "access", topdir_acl->dirname);

		    if ((user_acl->perm & dir_perm_modify) != dir_perm_modify)
			forbidden_in(line, "modify", topdir_acl->dirname);
		    break;

		case TOPDIRS_FILE:
		    if ((user_acl->perm & branch_perm_checkin) != branch_perm_checkin)
			forbidden_in(line, "checkin", topdir_acl->dirname);

		    if ((user_acl->perm & branch_perm_checkout) != branch_perm_checkout)
			forbidden_in(line, "checkout", topdir_acl->dirname);

		    break;
		}

		shrink_stringbuf(line, saved_len);
		user_acl = user_acl->next;
	    }
	}
    }
}

static int
print_out_acls_for_file(char* relative_name, char* filename, char* dirname, char* branch)
{
    int retval = 0;
    char* xbranch = branch ? branch : "trunk";
    struct stringbuf* line = NULL;
    struct stringbuf* fullname = NULL;
    struct DIR_ACL* diracl;
    struct BRANCH_ACL* branch_acl;
    struct BRANCH_ACE* file_ace;

    diracl = get_dir_acl(dirname, xbranch);
    if (!diracl)
	error(errno, 0, "Error getting directory ACLs for %s\n", dirname);

    line = new_stringbuf("/");
    if (!line)
	goto err;

    cat_stringbuf(line, filename);
    catc_stringbuf(line, '/');

    print_out_topdir_acls(diracl->top_dirs, TOPDIRS_FILE, NULL, line);

    branch_acl = get_branch_acl(diracl, xbranch);
    file_ace = get_file_on_branch_ace(branch_acl, filename);
    if (file_ace) {
	struct USER_ACL* user_acl = file_ace->user_acls;
	while (user_acl) {
	    int saved_len = line->len;

	    cat_stringbuf(line, user_acl->username);
	    cat_stringbuf(line, "/here/");

	    if ((user_acl->perm & branch_perm_checkin) != branch_perm_checkin)
		forbidden(line, "checkin");

	    if ((user_acl->perm & branch_perm_checkout) != branch_perm_checkout)
		forbidden(line, "checkout");

	    shrink_stringbuf(line, saved_len);

	    user_acl = user_acl->next;
	}
    }

    retval = 1;
 err:
    free_stringbuf(line);
    return retval;
}


static int
print_out_acls_for_dir(char* relative_name, char* branch)
{
    int retval = 0;
    int saved_len;
    struct stringbuf* fulldir = NULL;
    struct stringbuf* line = NULL;
    DIR_ACL* dir_acl = NULL;
    USER_ACL* user_acl;

    fulldir = get_fulldir(relative_name);
    if (!fulldir)
	goto out;

    line = new_stringbuf("DP/");
    if (!line)
	goto out;
    saved_len = line->len;

    if (fulldir->buf[fulldir->len - 1] == '/')
	fulldir->buf[fulldir->len - 1] = '\0';
	
    dir_acl = get_dir_acl(fulldir->buf, branch);
    if (!dir_acl)
	goto out;

    /* print out permissions denied in one of the top directories */
    print_out_topdir_acls(dir_acl->top_dirs, TOPDIRS_DIRECTORY, NULL, line);

    /* print out permissions denied in this directory */
    user_acl = dir_acl->user_acls;
    while (user_acl) {
	int saved_len = line->len;
	cat_stringbuf(line, user_acl->username);
	cat_stringbuf(line, "/here/");
	if ((user_acl->perm & dir_perm_access) != dir_perm_access)
	    forbidden(line, "access");

	if ((user_acl->perm & dir_perm_modify) != dir_perm_modify)
	    forbidden(line, "modify");

	shrink_stringbuf(line, saved_len);
	user_acl = user_acl->next;
    }

    retval = 1;
 out:
    free_stringbuf(fulldir);
    free_stringbuf(line);
    return retval;
}


static int
lsacl_fileproc (void *callerdat, struct file_info *finfo)
{
    print_out_acls_for_file(finfo->fullname, finfo->file, finfo->repository, tagname);

    return 0;
}

/* ARGSUSED */
static Dtype
lsacl_dirproc (callerdat, dir, repository, update_dir, entries)
    void *callerdat;
    char *dir;
    char *repository;
    char *update_dir;
    List *entries;
{
    /* Should we go this way? AM */
#ifdef SERVER_SUPPORT
    int grant = grant_dir_permission(repository, dir, dir_perm_access, "list");
#else
    int grant = 1;
#endif

    if (deep > 0) {
	cvs_output("D", 1);

	if (!skip_prefix)
	    cvs_output("/",1);
	
	/* Is it really safe. I mean is 'update_dir' is always
	 * startdir + smth?
	 */
	cvs_output(update_dir + skip_prefix, 0);
	if (!grant)
	    cvs_output("//locked",0);
	cvs_output("\n", 1);
    }
    deep++;
    print_out_acls_for_dir(update_dir, tagname);

#ifdef SERVER_SUPPORT
    if (!grant)
	return R_SKIP_ALL;
#endif
    if (local_only && deep > 1)
	return R_SKIP_ALL;
    return R_PROCESS;
}

static int 
lsacl_proc (int argc, char **argv, char *xwhere,
	 char *mwhere, char *mfile, int shorten,
	 int local, char *mname, char *msg)
{
    int err = 0;
    int which = W_REPOS | W_ATTIC;
    struct stringbuf* repository;

    skip_prefix = (strcmp(".",argv[1])) ? strlen(argv[1]) : 0;

    repository = new_stringbuf(current_parsed_root->directory);

    if (CVS_CHDIR(repository->buf) < 0) {
	err = 1;
	goto out;
    }
	
    deep = 0;
    local = 0; /* we always use recursion mode, 'cause we anyway need
		  subdirectores listing
	       */
    err = start_recursion (lsacl_fileproc, (FILESDONEPROC) NULL, lsacl_dirproc,
			   (DIRLEAVEPROC) NULL, (void *) NULL,
			   argc - 1, argv + 1, local, which, 0, 1,
			   "", 1);

 out:
    free_stringbuf(repository);
    return err;
}


static int
server_side_lsacl (int argc, char **argv)
{
    int err = 0;
    int i;

    for (i = 0; i < argc; i++) {
	err += lsacl_proc (argc + 1, argv - 1, (char *) NULL,
			(char *) NULL, (char *) NULL, 0, local_only, (char *) NULL,
			(char *) NULL);
    }
    return err;
}

int
lsacl (int argc, char** argv)
{
    int c;

    optind = 0;
    while ((c = getopt (argc, argv, "Rr:")) != -1)
    {
	switch (c)
	{
	case 'R':
	    local_only = 0;
	    break;

	case 'r':
	    tagname = xstrdup(optarg);
	    break;

	default:
	    usage(lsacl_usage);
	    break;
	}

    };

    argc -= optind;
    argv += optind;
    if (argc < 1 || !(*argv[0]))
	usage (lsacl_usage);
	

#ifdef CLIENT_SUPPORT
    if (current_parsed_root->isremote) {
	int i;

	start_server();

	ign_setup();

	if (!local_only)
	    send_to_server ("Argument -R\012", 0);

	if (tagname != NULL)
	    option_with_arg ("-r", tagname);

	for (i = 0; i < argc; i++)
	    send_arg (argv[i]);

	send_to_server("lsacl\012", 0);

	return get_responses_and_close();
    }
#endif /* CLIENT_SUPPORT */

    return server_side_lsacl(argc, argv);
}
