/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   Network client based on BSD sockets.

*/

#include "config.h"

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "config.h"
#include "root.h"

#include "network-client.h"
#include "socket-client.h"
#include "fd-client.h"
#include "http_proxy-client.h"

struct socket_client_data {
    int sockfd;
    char *hostname;
    short port;
};

typedef struct socket_client_data *SCD;

static int socket_connect(NCLI);
static int socket_disconnect(NCLI);

static int socket_read(NCLI, char *buf, unsigned int len);
static int socket_write(NCLI, char *buf, unsigned int len);
static int socket_flush(NCLI);

struct network_client *
create_socket_client (cvsroot_t *CVSROOT, ncli_error_cb error_cb)
{
    SCD scd;
    struct network_client *nc = malloc(sizeof(struct network_client));
    if (nc == NULL) {
	return NULL;
    }

    scd = malloc(sizeof(struct socket_client_data));
    if (scd == NULL) {
	return NULL;
    }

    scd->hostname = CVSROOT->hostname;
    scd->port = CVSROOT->port;

    nc->connect = socket_connect;
    nc->disconnect = socket_disconnect;

    nc->read = socket_read;
    nc->write = socket_write;
    nc->flush = socket_flush;

    nc->error = error_cb;

    nc->client_data = scd;

    return nc;
}

int
ncli_connect_socket (NCLI ncli, char *hostname, int port)
{
	if (ncli_need_http_proxy(ncli,hostname,port))
		return ncli_connect_http_proxy(ncli,hostname,port);
	else
		return ncli_connect_socket_direct(ncli,hostname,port);
}

int
ncli_connect_socket_direct (NCLI ncli, char *hostname, int port)
{
    struct hostent *hostinfo;
    struct sockaddr_in sai;

    /* create sending socket */
    int sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == -1)
	return ncli->error(ncli, "Cannot create socket: %s",
			   strerror(errno));

    /* initialize server connection info */
    memset(&sai, 0, sizeof sai);
    sai.sin_family = AF_INET;
    sai.sin_port = htons (port);

    /* resolve server hostname */
    hostinfo = gethostbyname(hostname);
    if (hostinfo == NULL)
	return ncli->error(ncli, "Cannot resolve hostname '%s': %s", 
			   hostname, strerror(errno));
    sai.sin_addr = *(struct in_addr *) hostinfo->h_addr;


    /* connect to server */
    if (connect(sock, (struct sockaddr*)&sai, sizeof(sai)) == -1)
	return ncli->error(ncli, "Cannot connect to %s:%d %s",
			   hostname, port, strerror(errno));

    /* set close-on-exec */
    if (fcntl(sock, F_SETFD, 1) == -1)
	return ncli->error(ncli, "Cannot set close-on-exec: %s",
			   strerror(errno));

    return sock;
}

static int
socket_connect (NCLI ncli)
{
    SCD scd = (SCD) ncli->client_data;

    scd->sockfd = ncli_connect_socket(ncli, scd->hostname, scd->port);

    return NCLI_SUCCESS;
}


static int
socket_disconnect (NCLI ncli)
{
    SCD scd = (SCD) ncli->client_data;

    if (shutdown(scd->sockfd, 2) == -1)
	return ncli->error(ncli, "Cannot shutdown connection to %s: %s",
			   scd->hostname, strerror(errno));

    return NCLI_SUCCESS;
}


static int
socket_read (NCLI ncli, char *buf, unsigned int len)
{
    SCD scd = (SCD) ncli->client_data;
    int ret;
    struct pollfd fds[1];

    /* wait for the activity on file descriptor */
    fds[0].fd = scd->sockfd;
    fds[0].events = POLLIN;

    if (poll(fds, 1, -1) == -1) {
	if ((errno == EAGAIN) || (errno == EINTR))
	    return -1;
	return ncli->error(ncli, "poll() error: %s",
			   strerror(errno));
    }

    /* set non-blocking mode */
    if (fcntl(scd->sockfd, F_SETFL, O_NONBLOCK) == -1)
	return ncli->error(ncli, "Cannot set non-blocking mode: %s",
			   strerror(errno));

    ret = basic_fd_read(ncli, scd->sockfd, buf, len);

    /* unset non-blocking mode */
    if (fcntl(scd->sockfd, F_SETFL, fcntl(scd->sockfd, F_GETFL) & ~O_NONBLOCK) == -1)
	return ncli->error(ncli, "Cannot unset non-blocking mode: %s",
			   strerror(errno));

    return ret;
}


static int
socket_write (NCLI ncli, char *buf, unsigned int len)
{
    SCD scd = (SCD) ncli->client_data;

    return basic_fd_write(ncli, scd->sockfd, buf, len);
}


static int
socket_flush (NCLI ncli)
{
    (void) ncli;
    return NCLI_SUCCESS;
}
