/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001
   Copyright (c) Alexey Morozov <morozov@novosoft.ru>, Novosoft, Inc. 2001

   Network client with OpenSSL support

*/

#include <stdlib.h>

#include "config.h"

#ifdef HAVE_OPENSSL
#include <openssl/err.h>
#include <openssl/ssl.h>
#endif

#include "network-client.h"
#include "socket-client.h"
#include "ssl-client.h"

struct ssl_client_data {
    char *hostname;
    int port;
    SSL *ssl;
};

typedef struct ssl_client_data *SSLCD;

static int ssl_connect(NCLI);
static int ssl_disconnect(NCLI);

static int ssl_read(NCLI, char *buf, int len);
static int ssl_write(NCLI, char *buf, int len);
static int ssl_flush(NCLI);


struct network_client *
create_ssl_client (cvsroot_t *CVSROOT,
		   ncli_error_cb error_cb)
{
    SSLCD sslcd;

    struct network_client *nc = malloc(sizeof(struct network_client));
    if (nc == NULL)
	return NULL;

    sslcd = malloc(sizeof(struct ssl_client_data));
    if (sslcd == NULL)
	return NULL;

    sslcd->hostname = CVSROOT->hostname;
    sslcd->port = CVSROOT->port;

    nc->client_data = sslcd;

    nc->connect = ssl_connect;
    nc->disconnect = ssl_disconnect;

    nc->read = ssl_read;
    nc->write = ssl_write;
    nc->flush = ssl_flush;

    nc->error = error_cb;

    SSL_library_init();
    SSL_load_error_strings();

    return nc;
}

int
ssl_connect (NCLI ncli)
{
    SSLCD sslcd = (SSLCD) ncli->client_data;
    int sockfd = ncli_connect_socket(ncli, sslcd->hostname, sslcd->port);
    SSL_CTX *ctx;
    SSL *ssl;
    X509 *server_cert;
    char *str;
    long x509_err;

    ctx = SSL_CTX_new(SSLv3_client_method());

    SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);

    ssl = SSL_new(ctx);
    if (ssl == NULL) {
	long ssl_err = ERR_get_error();
	return ncli->error(ncli, "Could not create new SSL session: %s",
			   ERR_reason_error_string(ssl_err));
    }

    SSL_set_fd(ssl, sockfd);

    if (!SSL_connect(ssl)) {
	long ssl_err = ERR_get_error();
	return ncli->error(ncli, "Could not connect to %s:%d via SSL: %s",
			   sslcd->hostname, sslcd->port,
			   ERR_reason_error_string(ssl_err));
    }

    x509_err = SSL_get_verify_result(ssl);
    if (x509_err != X509_V_OK) {
	return ncli->error(ncli, "Certificate verification failed: %s",
			   X509_verify_cert_error_string(x509_err));
    }

    server_cert = SSL_get_peer_certificate(ssl);
    if (server_cert == NULL) {
	long ssl_err = ERR_get_error();
	return ncli->error(ncli, "Could not get peer certificate: %s",
			   ERR_reason_error_string(ssl_err));
    }

    str = X509_NAME_oneline (X509_get_subject_name (server_cert), 0, 0);
    if (str == NULL) {
	long ssl_err = ERR_get_error();
	return ncli->error(ncli, "Could not get subject name from certificate: %s",
			   ERR_reason_error_string(ssl_err));
    }
    printf ("Certificate subject: %s\n", str);
    free (str);

    str = X509_NAME_oneline (X509_get_issuer_name (server_cert), 0, 0);
    if (str == NULL) {
	long ssl_err = ERR_get_error(); 
	return ncli->error(ncli, "Could not get issuer name from certificate: %s",
			   ERR_reason_error_string(ssl_err));
    }
    printf ("Certificate issuer: %s\n", str);
    free (str);

    X509_free(server_cert);


    sslcd->ssl = ssl;

    return NCLI_SUCCESS;
}


int
ssl_disconnect (NCLI ncli)
{
    SSLCD sslcd = (SSLCD) ncli->client_data;

    SSL_shutdown(sslcd->ssl);
    SSL_free(sslcd->ssl);

    return NCLI_SUCCESS;
}


int
ssl_read (NCLI ncli, char *buf, int len)
{
    SSLCD sslcd = (SSLCD) ncli->client_data;
    SSL *ssl = sslcd->ssl;
    int nread = 0;

    while (len > 0) {
	int n = SSL_read(ssl, buf + nread, len);

	long ssl_err = SSL_get_error(ssl, n);
	if (ssl_err == SSL_ERROR_ZERO_RETURN)
	    break;

	if (ssl_err != SSL_ERROR_NONE)
	    return ncli->error(ncli, "Could not read SSL data: %s",
			       ERR_reason_error_string(ssl_err));

	len -= n;
	nread += n;
    }

    return nread;
}

int
ssl_write (NCLI ncli, char *buf, int len)
{
    SSLCD sslcd = (SSLCD) ncli->client_data;
    SSL *ssl = sslcd->ssl;

    while (len > 0) {
	int n = SSL_write(ssl, buf, len);

	long ssl_err = SSL_get_error(ssl, n);
	if (ssl_err == SSL_ERROR_ZERO_RETURN)
	    break;

	if (ssl_err != SSL_ERROR_NONE)
	    return ncli->error(ncli, "Could not write SSL data: %s",
			       ERR_reason_error_string(ssl_err));

	len -= n;
	buf += n;
    }

    return NCLI_SUCCESS;
}

int ssl_flush (NCLI ncli)
{

    return NCLI_SUCCESS;
}
