/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001-2002

   Network client using transparent gzip compression.

*/

#include "zlib-client.h"

#include "error.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#undef EXTRA_VERBOSE

struct zlib_client_data {
    Bytef buf_in[BUFSIZ];
    z_stream zsin;
    Bytef buf_out[BUFSIZ];
    z_stream zsout;
    int inflate_initialized;
    int zsin_stream_end;
};

typedef struct zlib_client_data *ZCD;


static NCLI_STATUS zlib_connect(NCLI);
static NCLI_STATUS zlib_disconnect(NCLI);

static int zlib_read(NCLI, char *buf, unsigned int len);
static NCLI_STATUS zlib_write(NCLI, char *buf, unsigned int len);
static NCLI_STATUS zlib_flush(NCLI);

static int gzip_level = 0;

static int file_gzip_level = 0;

struct network_client *
create_zlib_client (cvsroot_t *CVSROOT,
		    ncli_error_cb error_cb)
{
    ZCD zcd;
    struct network_client *nc = malloc(sizeof(struct network_client));
    if (!nc)
	return NULL;

    (void) CVSROOT;

    zcd = malloc(sizeof(struct zlib_client_data));
    if (!zcd)
	return NULL;

    zcd->inflate_initialized = 0;

    nc->connect = zlib_connect;
    nc->disconnect = zlib_disconnect;

    nc->read = zlib_read;
    nc->write = zlib_write;
    nc->flush = zlib_flush;

    nc->error = error_cb;
    nc->client_data = zcd;

    return nc;
}


static NCLI_STATUS
zlib_connect(NCLI ncli)
{
    ZCD zcd = (ZCD) ncli->client_data;
    int ret;

#ifdef EXTRA_VERBOSE
    fprintf(stderr, "zlib_connect entered\n");
#endif

    zcd->zsout.zalloc = Z_NULL;
    zcd->zsout.zfree = Z_NULL;
    zcd->zsout.opaque = Z_NULL;
    ret = deflateInit(&( zcd->zsout ), 5);

#ifdef EXTRA_VERBOSE
    fprintf(stderr, "zlib_connect: deflateInit() = %d\n", ret);
#endif

    zcd->zsin.zalloc = Z_NULL;
    zcd->zsin.zfree = Z_NULL;
    zcd->zsin.opaque = Z_NULL;

    return NCLI_SUCCESS;
}


static NCLI_STATUS
zlib_disconnect(NCLI ncli)
{
    ZCD zcd = (ZCD) ncli->client_data;
    int ret;
    Byte buf[BUFSIZ];

    for (;;) {
	int err;
	zcd->zsout.next_out = buf;
	zcd->zsout.avail_out = sizeof(buf);
	err = deflate(&( zcd->zsout ), Z_FINISH);
	if (err == Z_STREAM_END)
	    break;
	if (err != Z_OK)
	    return NCLI_FAILURE;
#ifdef EXTRA_VERBOSE
	fprintf(stderr, "Z_FINISH: writing %d bytes\n", zcd->zsout.next_out - buf);
#endif
	ncli->next->write(ncli->next, buf, zcd->zsout.next_out - buf);
    }

    ret = deflateEnd(&( zcd->zsout ));
#ifdef EXTRA_VERBOSE
    fprintf(stderr, "zlib_disconnect: deflateEnd() = %d\n", ret);
#endif

    if (zcd->inflate_initialized) {
	ret = inflateEnd(&( zcd->zsin ));
#ifdef EXTRA_VERBOSE
	fprintf(stderr, "zlib_disconnect: inflateEnd() = %d\n", ret);
#endif
    }

    return NCLI_SUCCESS;
}


static int
zlib_inflate_more(NCLI ncli, unsigned int len)
{
    ZCD zcd = (ZCD) ncli->client_data;
    int ret;
    int uncompr_len;

    if (!zcd->inflate_initialized) {
	ret = inflateInit(&( zcd->zsin ));
	if (ret != Z_OK)
	    return ncli->error(ncli, "inflateInit(): %s", zcd->zsin.msg);

	zcd->inflate_initialized = 1;

#ifdef EXTRA_VERBOSE
	fprintf(stderr, "after inflateInit: zsin.avail_in = %d; next_in = %d\n",
		zcd->zsin.avail_in, zcd->zsin.next_in - zcd->buf_in);
#endif
    }

    ret = inflate(&( zcd->zsin ), Z_NO_FLUSH);
    uncompr_len = len - zcd->zsin.avail_out;
#ifdef EXTRA_VERBOSE
    fprintf(stderr, "uncompr_len == %d\n", uncompr_len);
#endif
    switch (ret) {
    case Z_STREAM_END:
#ifdef EXTRA_VERBOSE
	fprintf(stderr, "Setting zcd->zsin_stream_end = 1\n");
#endif
	zcd->zsin_stream_end = 1;
	/* fall through */

    case Z_OK:
#ifdef EXTRA_VERBOSE
	fprintf(stderr, "inflate_more: uncompr_len = %d\n", uncompr_len);
#endif
	if (uncompr_len)
	    return uncompr_len;
	else {
	    errno = EAGAIN;
	    return -1;
	}

    default:
	return ncli->error(ncli, "inflate(): %s", zcd->zsin.msg);
    }
}


static int
zlib_read(NCLI ncli, char *buf, unsigned int len)
{
    ZCD zcd = (ZCD) ncli->client_data;
    int nread;

    if (zcd->zsin_stream_end)
	return 0;

    zcd->zsin.next_out = buf;
    zcd->zsin.avail_out = len;

    do {
	int len_to_read = len > sizeof(zcd->buf_in) ? sizeof(zcd->buf_in) : len;
	int uncompr_len;
	/*
	  if there are data in the zsin buffer, we uncompress it (no more than
	  requested) and return;
	*/
#ifdef EXTRA_VERBOSE
	fprintf(stderr, "zcd->zsin.avail_in == %d\n", zcd->zsin.avail_in);
#endif
	if (zcd->zsin.avail_in > 0)
	    return zlib_inflate_more(ncli, len);

	/* 
	   otherwise we do the read(), passing it either the same LEN
	   as requested(that's just a heuristics), but no more than
	   the length of buffer;
	*/
	zcd->zsin.next_in = zcd->buf_in;
	nread = ncli->next->read(ncli->next, zcd->buf_in, len_to_read);
#ifdef EXTRA_VERBOSE
	fprintf(stderr, "nread == %d\n", nread);
#endif
	/*uncompr_len = len_to_read - zcd->zsin.avail_out;*/
	uncompr_len = len - zcd->zsin.avail_out;

	/*
	  if read() returns -1 && errno == EAGAIN: if we have some
	  uncompressed data, return those; otherwise, return -1 &&
	  errno == EAGAIN;
	*/
	if ((nread == -1) && (errno == EAGAIN)) {
	    if (uncompr_len)
		return uncompr_len;
	    else
		return -1;
	}

	/*
	  if read() returns -1, we return -1;
	*/
	if (nread == -1)
	    return ncli->error(ncli, "read() error: %s", strerror(errno));

	/*
	  if read() returns 0, that means end of file and we must
	  return what we have read so far (maybe zero, too, if client
	  attempted to read again);
	*/
	if (nread == 0)
	    return uncompr_len;
	
	/*
	  otherwise, if read() returns > 0, we go to the beginning;
	*/
	zcd->zsin.avail_in = nread;


    } while (1);

}


static NCLI_STATUS
zlib_write(NCLI ncli, char *buf, unsigned int len)
{
    ZCD zcd = (ZCD) ncli->client_data;
    int ret;

#ifdef EXTRA_VERBOSE
    fprintf(stderr, "zlib_write(%d)\n", len);
#endif

    zcd->zsout.next_in = buf;
    zcd->zsout.avail_in = len;
  do {
    zcd->zsout.next_out = zcd->buf_out;
    zcd->zsout.avail_out = BUFSIZ;

    ret = deflate(&( zcd->zsout ), Z_SYNC_FLUSH);
    if (ret != Z_OK)
	return NCLI_FAILURE;

#ifdef EXTRA_VERBOSE
    fprintf(stderr, "deflate() == %d\n", ret);
    fprintf(stderr, "z_write: zsout: total_in = %ld; total_out = %ld; avail_out = %d\n", 
	   zcd->zsout.total_in, zcd->zsout.total_out, zcd->zsout.avail_out);

    fprintf(stderr, "Actually writing %d bytes\n", BUFSIZ - zcd->zsout.avail_out);
#endif
    ret = ncli->next->write(ncli->next, zcd->buf_out, BUFSIZ - zcd->zsout.avail_out);
    if (ret != NCLI_SUCCESS)
        return ret;
  } while(zcd->zsout.avail_out == 0);
  return ret;
}




static NCLI_STATUS
zlib_flush(NCLI ncli)
{
    (void) ncli;
    return NCLI_SUCCESS;
}


int
get_gzip_level (void)
{
    return gzip_level;
}

void
set_gzip_level (int level)
{
    if (level < 0 || level > 9)
	error (1, 0,
	       "gzip compression level must be between 0 and 9");

    gzip_level = level;
}


int
get_file_gzip_level (void)
{
    return file_gzip_level;
}

void
set_file_gzip_level (int level)
{
    file_gzip_level = level;
}


